// amp11 - an Audio-MPEG decoder - sample
// Copyright (c) 1997 Niklas Beisert
// See COPYING (GNU General Public License 2) for license

#define VERSION "nb221197"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef __DOS4G__
#define KBBREAK
#include <conio.h>
#include "timer.h"
#define CLOCKGRAN 1000
#define getclock tmGetTimer
#define initclock tmInit
#define closeclock tmClose
#include "binfplsb.h"
#endif

#ifdef __NT__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define CLOCKGRAN 1000
#define getclock timeGetTime
#define initclock()
#define closeclock()
#define KBBREAK
#include <conio.h>
#include "binfplnt.h"
#endif

#ifndef CLOCKGRAN
#include <time.h>
#ifdef CLOCKS_PER_SEC
#define CLOCKGRAN CLOCKS_PER_SEC
#else
#define CLOCKGRAN 1000000
#endif
#define getclock clock
#define initclock()
#define closeclock()
#endif

#ifdef LINUX
#include "binfpllx.h"
#endif

#include "binfhttp.h"
#include "binfstd.h"
#include "binfarc.h"
#include "binfplwv.h"
#include "binfcon.h"
#include "mpxdec.h"

int main(int argc, char **argv)
{
  char *proxy=0;
  char *filename=0;
  char *outname=0;
  char *logname=0;
  int i;
  float vol=1;
  float bal=0;
  int tonull=0;
  int down=0;
  int chn=0;
  int quiet=0;
  float ctr=0;
  float sep=1;
  float srnd=1;
  float pitch=0;
  float pitch0=1764;
  for (i=1; i<argc; i++)
    if (argv[i][0]=='-')
    {
      switch (argv[i][1])
      {
      case 'n': tonull=1; break;
      case 'm': chn=1; break;
      case 'q': quiet=1; break;
      case 's': chn=2; break;
      case 'S': chn=-2; break;
      case '2': down=1; break;
      case '4': down=2; break;
      case 'i': srnd=-1; break;
      case 'P': proxy=argv[i]+2; break;
      case 'l': logname=argv[i]+2; break;
      case 'v': vol=atof(argv[i]+2); break;
      case 'b': bal=atof(argv[i]+2); break;
      case 'c': ctr=atof(argv[i]+2); break;
      case 'p': sep=atof(argv[i]+2); break;
      case 'f': pitch=atof(argv[i]+2); break;
      case 'F': pitch0=atof(argv[i]+2); break;
      case 'o': outname=argv[i]+2; break;
      default: printf("unknown switch\n"); return 1;
      }
    }
    else
    {
      if (filename)
      {
        printf("too many files\n");
        return 1;
      }
      else
        filename=argv[i];
    }

  if (!quiet)
    printf("amp11   Audio-MPEG decoder (%s)  (c) '97 Niklas Beisert\n\n", VERSION);

  if (!filename)
  {
    printf("amp11 comes with ABSOLUTELY NO WARRANTY, for details see COPYING (GPL 2)\n");
    printf("\n");
    printf("amp11 [-q] [-n] [-PPROXY] [-lLOG] INFILE|URL [-oOUTFILE]\n");
    printf("      [-m] [-s] [-S] [-2] [-4]\n");
    printf("      [-vVOL]\n");
    printf("      [-pSEP] [-cCTR] [-bBAL] [-i]\n");
    printf("      [-fPITCH] [-FFREQ0]\n");
    return 1;
  }

  binfile *musicin=0;

  sbinfile smusicin;
  httpbinfile httpmusicin;
  if (!strncmp(filename, "http://", 7))
  {
    if (httpmusicin.open(filename, proxy, 0, 0))
    {
      printf("Could not open input URL\n");
      return 1;
    }
    httpmusicin.ioctl(httpmusicin.ioctlrrbufset, 16384);
    musicin=&httpmusicin;
  }
  else
  {
    if (smusicin.open(filename, sbinfile::openro)<0)
    {
      printf("Could not open input file\n");
      return 1;
    }

    musicin=&smusicin;
  }

  sbinfile logfile;
  if (logname)
  {
    logfile.open(logname, sbinfile::openrw|sbinfile::opentr);
    musicin->ioctl(musicin->ioctlrsetlog, &logfile, 0);
  }

  abinfile awavedata;
  binfile *wavedata=musicin;
  char riff[4];
  musicin->peek(riff, 4);
  if (!memcmp(riff, "RIFF", 4))
  {
    int blklen=12;
    while (blklen&&!musicin->ioctl(musicin->ioctlreof))
      blklen-=musicin->read(0, blklen);
    while (1)
    {
      if (musicin->read(riff, 4)!=4)
      {
        printf("Could not find data chunk in RIFF file\n");
        return 1;
      }
      blklen=getil4(*musicin);
      if (!memcmp(riff, "data", 4))
        break;
      while (blklen&&!musicin->ioctl(musicin->ioctlreof))
        blklen-=musicin->read(0, blklen);
    }
    awavedata.open(*musicin, musicin->tell(), blklen);
    wavedata=&awavedata;
  }
  else
  if (musicin->getmode()&binfile::modeseek)
  {
    musicin->seekend(-128);
    musicin->read(riff, 3);
    awavedata.open(*musicin, 0, musicin->length()-(memcmp(riff, "TAG", 3)?0:128));
    wavedata=&awavedata;
  }

  int freq,stereo;

  ampegdecoder dec;
  if (dec.open(*wavedata, freq, stereo, 1, down, chn))
  {
    printf("Could not open audio mpeg\n");
    return 1;
  }

  binfile *outfile;
  sbinfile diskfile;
  wavplaybinfile wavfile;
  binfile nofile;
  outfile=&nofile;

#ifdef __NT__
  ntplaybinfile musicout;
#endif
#ifdef __DOS4G__
  sbplaybinfile musicout;
#endif
#ifdef LINUX
  linuxplaybinfile musicout;
#endif

  if (outname)
  {
    if (diskfile.open(outname, sbinfile::openrw|sbinfile::opentr))
    {
      printf("Could not create output file\n");
      return 1;
    }
    outfile=&diskfile;
    if (!strcmp(outname+strlen(outname)-4, ".wav"))
    {
      wavfile.open(diskfile, freq, stereo, 1, 0);
      outfile=&wavfile;
    }
  }
#ifdef __NT__
  else
  if (!tonull)
  {
    if (musicout.open(freq, stereo, 1, 4608*4, 32))
    {
      printf("Could not create output stream\n");
      return 1;
    }
    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
    outfile=&musicout;
  }
#endif
#ifdef LINUX
  else
  if (!tonull)
  {
    if (musicout.open(freq, stereo, 1))
    {
      printf("Could not create output stream\n");
      return 1;
    }
//    setpriority(PRIO_PROCESS,getpid(),-20);
    outfile=&musicout;
  }
#endif
#ifdef __DOS4G__
  else
  if (!tonull)
  {
    if (musicout.open(freq, stereo, 1, 0x20000))
    {
      printf("Could not create output stream\n");
      return 1;
    }
    outfile=&musicout;
  }
#endif

  float vols[3][3];
  vols[0][0]=0.5*(1.0-bal)*(1.0-ctr+sep);
  vols[0][1]=0.5*(1.0-bal)*(1.0+ctr-sep);
  vols[0][2]=1-bal;
  vols[1][0]=0.5*(1.0+bal)*(1.0-ctr-sep)*srnd;
  vols[1][1]=0.5*(1.0+bal)*(1.0+ctr+sep)*srnd;
  vols[1][2]=(1.0+bal)*srnd;
  vols[2][0]=0.5*(1.0-ctr);
  vols[2][1]=0.5*(1.0+ctr);
  vols[2][2]=1.0;
  dec.ioctl(dec.ioctlsetstereo, vols, 0);
  vols[0][0]=vol;
  dec.ioctl(dec.ioctlsetvol, &vols[0][0], 0);

  static float l3equal[576];
  for (i=0; i<576; i++)
    l3equal[i]=exp(pitch*log((i+0.5)*freq/(576.0*2*pitch0)));
  dec.ioctl(dec.ioctlsetequal576, l3equal, 0);

  short sampbuf[2304];

#ifdef TIME
  initclock();
  long t0=0;
#endif

  peekch(dec);

  while (1)
  {
#ifdef KBBREAK
    if (kbhit())
      break;
#endif
#ifdef TIME
    if (!quiet)
      t0-=getclock();
#endif
    int l=dec.read(sampbuf, 4608);
#ifdef TIME
    if (!quiet)
      t0+=getclock();
#endif
    if (!l)
      break;
    outfile->write(sampbuf, l);
    if (!quiet)
    {
      fprintf(stdout, "\r%.3f", (float)((float)dec.tell()/((stereo?4:2)*freq)));
      if (dec.length())
        fprintf(stdout, "/%.3f=%.3f", (float)((float)dec.length()/((stereo?4:2)*freq)), (float)((float)dec.tell()/dec.length()));
#ifdef TIME
      float t=(float)t0/CLOCKGRAN;
      fprintf(stdout, ":%.3f=%.3f", (float)t, (float)((float)t*(stereo?4:2)*freq/dec.tell()));
#endif
      fflush(stdout);
    }
  }

#ifdef TIME
  closeclock();
#endif

#ifdef KBBREAK
  while (kbhit())
    getch();
#endif

  if (!quiet)
    printf("\n");

  dec.close();
  wavedata->close();
  musicin->close();
  logfile.close();
  outfile->close();
  diskfile.close();

  return 0;
}
