// amp11 - an Audio-MPEG decoder - layer 1 decoder
// Copyright (c) 1997 Niklas Beisert
// See COPYING (GNU General Public License 2) for license

#include <math.h>
#include "mp1dec.h"

float ampegdecoderl1::rangefac[16];
float ampegdecoderl1::multiple[64];

int ampegdecoderl1::decode(void *outsamp)
{
  if (!decodehdr())
    return 0;

  int bitsleft=12*1000*ratetab[hdrversion?1:0][2][hdrbitrate]/samprate*32;
  if (hdrpadding)
    bitsleft+=32;
  bitsleft-=32;
  if (hdrcrc)
    bitsleft-=16;

  synthsetbuf(outsamp);

  int jsbound=(hdrmode==1)?(hdrmodeext+1)*4:(hdrmode==3)?0:32;
  int stereo = (hdrmode==3)?1:2;

  int i,j,q;

  for (i=0;i<32;i++)
    for (j=0;j<((i<jsbound)?2:1);j++)
    {
      bitsleft-=4;
      bitalloc[j][i] = mpgetbits(4);
      if (i>=jsbound)
        bitalloc[1][i] = bitalloc[0][i];
    }

  for (i=0;i<32;i++)
    for (j=0;j<stereo;j++)
      if (bitalloc[j][i])
      {
        bitsleft-=6;
        scale[j][i]=multiple[mpgetbits(6)]*rangefac[bitalloc[j][i]];
      }

  for (q=0;q<12;q++)
  {
    for (i=0;i<32;i++)
      for (j=0;j<((i<jsbound)?2:1);j++)
        if (bitalloc[j][i])
        {
          bitsleft-=bitalloc[j][i]+1;
          int s=mpgetbits(bitalloc[j][i]+1)-(1<<bitalloc[j][i])+1;
          fraction[j][i]=scale[j][i]*s;
          if (i>=jsbound)
            fraction[1][i]=scale[1][i]*s;
        }
        else
        {
          fraction[j][i]=0;
          if (i>=jsbound)
            fraction[1][i]=0;
        }
    synth(fraction[0]);
  }
  mpgetbits(bitsleft);
  return 1;
}

int ampegdecoderl1::openlayer(int, int rate, int freq, int stereo)
{
  if (orglay!=0)
    return 0;
  if (!stream)
  {
    slotsize=4;
    slotdiv=freq;
    nslots=(12*rate)/freq;
    fslots=(12*rate)%slotdiv;
    seekinitframes=2;
  }
  framesize=384*(stereo?2:1);

  int i;
  for (i=0; i<63; i++)
    multiple[i]=exp(log(2)*(3-i)/3.0);
  multiple[63]=0;
  for (i=0; i<16; i++)
    rangefac[i]=2.0/((2<<i)-1);
  return 1;
}

void ampegdecoderl1::closelayer()
{
}

void ampegdecoderl1::seekinit(int discard)
{
  synthsetbuf(0);
  int i,j;
  for (j=0; j<discard; j++)
    for (i=0; i<12; i++)
      synth(0);
  for (j=discard; j<2; j++)
    decode(0);
}

ampegdecoderl1::ampegdecoderl1()
{
}

ampegdecoderl1::~ampegdecoderl1()
{
}

