// amp11 - an Audio-MPEG decoder - layer 2 decoder
// Copyright (c) 1997 Niklas Beisert
// See COPYING (GNU General Public License 2) for license

#include <math.h>
#include "mp2dec.h"

ampegdecoderl2::sballoc ampegdecoderl2::alloc[17];

ampegdecoderl2::sballoc *ampegdecoderl2::atab0[]=
{
  0        , &alloc[0], &alloc[3], &alloc[4], &alloc[5], &alloc[6], &alloc[7], &alloc[8],
  &alloc[9], &alloc[10],&alloc[11],&alloc[12],&alloc[13],&alloc[14],&alloc[15],&alloc[16]
};

ampegdecoderl2::sballoc *ampegdecoderl2::atab1[]=
{
  0        , &alloc[0], &alloc[1], &alloc[3], &alloc[2], &alloc[4], &alloc[5], &alloc[6],
  &alloc[7], &alloc[8], &alloc[9], &alloc[10],&alloc[11],&alloc[12],&alloc[13],&alloc[16]
};

ampegdecoderl2::sballoc *ampegdecoderl2::atab2[]=
{
  0        , &alloc[0], &alloc[1], &alloc[3], &alloc[2], &alloc[4], &alloc[5], &alloc[16]
};

ampegdecoderl2::sballoc *ampegdecoderl2::atab3[]=
{
  0        , &alloc[0], &alloc[1], &alloc[16]
};

ampegdecoderl2::sballoc *ampegdecoderl2::atab4[]=
{
  0        , &alloc[0], &alloc[1], &alloc[2], &alloc[4], &alloc[5], &alloc[6], &alloc[7],
  &alloc[8], &alloc[9], &alloc[10],&alloc[11],&alloc[12],&alloc[13],&alloc[14],&alloc[15]
};

const int ampegdecoderl2::alloctablens[5]={27,30,8,12,30};

ampegdecoderl2::sballoc **ampegdecoderl2::alloctabs[3][32]=
{
  {
    atab0, atab0, atab0, atab1, atab1, atab1, atab1, atab1, atab1, atab1,
    atab1, atab2, atab2, atab2, atab2, atab2, atab2, atab2, atab2, atab2,
    atab2, atab2, atab2, atab3, atab3, atab3, atab3, atab3, atab3, atab3
  },
  {
    atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4,
    atab4, atab4
  },
  {
    atab1, atab1, atab1, atab1, atab4, atab4, atab4, atab4, atab4, atab4,
    atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4,
    atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4, atab4
  }
};

const int ampegdecoderl2::alloctabbits[3][32]=
{
  {4,4,4,4,4,4,4,4,4,4,4,3,3,3,3,3,3,3,3,3,3,3,3,2,2,2,2,2,2,2},
  {4,4,3,3,3,3,3,3,3,3,3,3},
  {4,4,4,4,3,3,3,3,3,3,3,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2}
};

float ampegdecoderl2::multiple[64];

int ampegdecoderl2::decode(void *outsamp)
{
  if (!decodehdr())
    return 0;

  int bitsleft=144*1000*ratetab[hdrversion?1:0][2][hdrbitrate]/samprate*8;
  if (hdrpadding)
    bitsleft+=8;
  bitsleft-=32;
  if (hdrcrc)
    bitsleft-=16;

  synthsetbuf(outsamp);

  int stereo = (hdrmode==3)?1:2;
  int brpch = ratetab[0][1][hdrbitrate] / stereo;

  int tabnum;
  if (hdrversion)
    tabnum=4;
  else
  if (((hdrfreq==1)&&(brpch>=56))||((brpch>=56)&&(brpch<=80)))
    tabnum=0;
  else
  if ((hdrfreq!=1)&&(brpch>=96))
    tabnum=1;
  else
  if ((hdrfreq!=2)&&(brpch<=48))
    tabnum=2;
  else
    tabnum=3;
  sballoc ***alloc=alloctabs[tabnum>>1];
  const int *allocbits=alloctabbits[tabnum>>1];
  int sblimit=alloctablens[tabnum];
  int jsbound;
  if (hdrmode==1)
    jsbound=(hdrmodeext+1)*4;
  else
  if (hdrmode==3)
    jsbound=0;
  else
    jsbound=sblimit;

  int i,j,k,q;

  for (i=0; i<sblimit; i++)
    for (j=0; j<((i<jsbound)?2:1); j++)
    {
      bitsleft-=allocbits[i];
      bitallocx[i][j]=alloc[i][mpgetbits(allocbits[i])];
      if (i>=jsbound)
        bitallocx[i][1]=bitallocx[i][0];
    }

  for (i=0; i<sblimit; i++)
    for (j=0; j<stereo; j++)
      if (bitallocx[i][j])
      {
        bitsleft-=2;
        scfsi[i][j]=mpgetbits(2);
      }

  for (i=0;i<sblimit;i++)
    for (j=0;j<stereo;j++)
      if (bitallocx[i][j])
      {
        int si[3];
        switch (scfsi[i][j])
        {
        case 0:
          bitsleft-=18;
          si[0]=mpgetbits(6);
          si[1]=mpgetbits(6);
          si[2]=mpgetbits(6);
          break;
        case 1:
          bitsleft-=12;
          si[0]=si[1]=mpgetbits(6);
          si[2]=mpgetbits(6);
          break;
        case 2:
          bitsleft-=6;
          si[0]=si[1]=si[2]=mpgetbits(6);
          break;
        case 3:
          bitsleft-=12;
          si[0]=mpgetbits(6);
          si[1]=si[2]=mpgetbits(6);
          break;
        }
        for (k=0; k<3; k++)
          scale[k][i][j]=multiple[si[k]]*bitallocx[i][j]->scale;
      }

  for (q=0;q<12;q++)
  {
    float (*scalex)[2]=scale[q>>2];

    for (i=0; i<sblimit; i++)
      for (j=0; j<((i<jsbound)?2:1); j++)
        if (bitallocx[i][j])
        {
          int s[3];
          if (!bitallocx[i][j]->steps)
            for (k=0; k<3; k++)
            {
              bitsleft-=bitallocx[i][j]->bits;
              s[k]=mpgetbits(bitallocx[i][j]->bits)-bitallocx[i][j]->scaleadd;
            }
          else
          {
            bitsleft-=bitallocx[i][j]->bits;
            int nlevels=bitallocx[i][j]->steps;
            int c=mpgetbits(bitallocx[i][j]->bits);
            for (k=0; k<3; k++)
            {
              s[k]=(c%nlevels)-bitallocx[i][j]->scaleadd;
              c/=nlevels;
            }
          }

          for (k=0; k<3; k++)
            fraction[k][j][i]=s[k]*scalex[i][j];
          if (i>=jsbound)
            for (k=0;k<3;k++)
              fraction[k][1][i]=s[k]*scalex[i][1];
        }
        else
        {
          for (k=0;k<3;k++)
            fraction[k][j][i]=0;
          if (i>=jsbound)
            for (k=0;k<3;k++)
              fraction[k][1][i]=0;
        }

    for (i=sblimit; i<32; i++)
      for (k=0; k<3; k++)
        for (j=0; j<stereo; j++)
          fraction[k][j][i]=0;

    for (k=0;k<3;k++)
      synth(fraction[k][0]);
  }
  mpgetbits(bitsleft);
  return 1;
}

int ampegdecoderl2::openlayer(int, int rate, int freq, int stereo)
{
  if (orglay!=1)
    return 0;
  if (!stream)
  {
    slotsize=1;
    slotdiv=freq;
    nslots=(144*rate)/freq;
    fslots=(144*rate)%slotdiv;
    seekinitframes=1;
  }
  framesize=1152*(stereo?2:1);

  int i;
  for (i=0; i<63; i++)
    multiple[i]=exp(log(2)*(3-i)/3.0);
  multiple[63]=0;
  for (i=0; i<3; i++)
  {
    alloc[i].scaleadd=1<<i;
    alloc[i].steps=2*alloc[i].scaleadd+1;
    alloc[i].bits=(10+5*i)>>1;
    alloc[i].scale=2.0/(2*alloc[i].scaleadd+1);
  }
  for (i=3; i<17; i++)
  {
    alloc[i].steps=0;
    alloc[i].bits=i;
    alloc[i].scaleadd=(1<<(i-1))-1;
    alloc[i].scale=2.0/(2*alloc[i].scaleadd+1);
  }
  return 1;
}

void ampegdecoderl2::closelayer()
{
}

void ampegdecoderl2::seekinit(int discard)
{
  synthsetbuf(0);
  int i;
  if (discard)
    for (i=0; i<16; i++)
      synth(0);
  else
    decode(0);
}

ampegdecoderl2::ampegdecoderl2()
{
}

ampegdecoderl2::~ampegdecoderl2()
{
}

